'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.applyOperation = applyOperation;

var _debug = require('debug');

var _debug2 = _interopRequireDefault(_debug);

var _warn = require('../utils/warn');

var _warn2 = _interopRequireDefault(_warn);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Debug.
 *
 * @type {Function}
 */

var debug = (0, _debug2.default)('slate:operation');

/**
 * Operations.
 *
 * @type {Object}
 */

var OPERATIONS = {
  // Text operations.
  insert_text: insertText,
  remove_text: removeText,
  // Mark operations.
  add_mark: addMark,
  remove_mark: removeMark,
  set_mark: setMark,
  // Node operations.
  insert_node: insertNode,
  join_node: joinNode,
  move_node: moveNode,
  remove_node: removeNode,
  set_node: setNode,
  split_node: splitNode,
  // Selection operations.
  set_selection: setSelection
};

/**
 * Apply an `operation` to the current state.
 *
 * @param {Transform} transform
 * @param {Object} operation
 * @return {Transform}
 */

function applyOperation(transform, operation) {
  var state = transform.state,
      operations = transform.operations;
  var type = operation.type;

  var fn = OPERATIONS[type];

  if (!fn) {
    throw new Error('Unknown operation type: "' + type + '".');
  }

  debug(type, operation);

  transform.state = fn(state, operation);
  transform.operations = operations.concat([operation]);

  return transform;
}

/**
 * Add mark to text at `offset` and `length` in node by `path`.
 *
 * @param {State} state
 * @param {Object} operation
 * @return {State}
 */

function addMark(state, operation) {
  var path = operation.path,
      offset = operation.offset,
      length = operation.length,
      mark = operation.mark;
  var _state = state,
      document = _state.document;

  var node = document.assertPath(path);
  node = node.addMark(offset, length, mark);
  document = document.updateDescendant(node);
  state = state.merge({ document: document });
  return state;
}

/**
 * Insert a `node` at `index` in a node by `path`.
 *
 * @param {State} state
 * @param {Object} operation
 * @return {State}
 */

function insertNode(state, operation) {
  var path = operation.path,
      index = operation.index,
      node = operation.node;
  var _state2 = state,
      document = _state2.document;

  var parent = document.assertPath(path);
  var isParent = document == parent;
  parent = parent.insertNode(index, node);
  document = isParent ? parent : document.updateDescendant(parent);
  state = state.merge({ document: document });
  return state;
}

/**
 * Insert `text` at `offset` in node by `path`.
 *
 * @param {State} state
 * @param {Object} operation
 * @return {State}
 */

function insertText(state, operation) {
  var path = operation.path,
      offset = operation.offset,
      text = operation.text,
      marks = operation.marks;
  var _state3 = state,
      document = _state3.document,
      selection = _state3.selection;
  var _selection = selection,
      startKey = _selection.startKey,
      endKey = _selection.endKey,
      startOffset = _selection.startOffset,
      endOffset = _selection.endOffset;

  var node = document.assertPath(path);

  // Update the document
  node = node.insertText(offset, text, marks);
  document = document.updateDescendant(node);

  // Update the selection
  if (startKey == node.key && startOffset >= offset) {
    selection = selection.moveStartOffset(text.length);
  }
  if (endKey == node.key && endOffset >= offset) {
    selection = selection.moveEndOffset(text.length);
  }

  state = state.merge({ document: document, selection: selection });
  return state;
}

/**
 * Join a node by `path` with a node `withPath`.
 *
 * @param {State} state
 * @param {Object} operation
 *   @param {Boolean} operation.deep (optional) Join recursively the
 *   respective last node and first node of the nodes' children. Like a zipper :)
 * @return {State}
 */

function joinNode(state, operation) {
  var path = operation.path,
      withPath = operation.withPath,
      _operation$deep = operation.deep,
      deep = _operation$deep === undefined ? false : _operation$deep;
  var _state4 = state,
      document = _state4.document,
      selection = _state4.selection;

  var first = document.assertPath(withPath);
  var second = document.assertPath(path);

  // Update doc
  document = document.joinNode(first, second, { deep: deep });

  // Update selection if we merged two texts together
  if (deep || second.kind == 'text') {
    var firstText = deep ? first.getLastText() : first;
    var secondText = deep ? second.getFirstText() : second;

    var _selection2 = selection,
        anchorKey = _selection2.anchorKey,
        anchorOffset = _selection2.anchorOffset,
        focusKey = _selection2.focusKey,
        focusOffset = _selection2.focusOffset;
    // The final key is the `first` key

    if (anchorKey == secondText.key) {
      selection = selection.merge({
        anchorKey: firstText.key,
        anchorOffset: anchorOffset + firstText.characters.size
      });
    }
    if (focusKey == secondText.key) {
      selection = selection.merge({
        focusKey: firstText.key,
        focusOffset: focusOffset + firstText.characters.size
      });
    }
  }

  state = state.merge({ document: document, selection: selection });
  return state;
}

/**
 * Move a node by `path` to a new parent by `path` and `index`.
 *
 * @param {State} state
 * @param {Object} operation
 * @return {State}
 */

function moveNode(state, operation) {
  var path = operation.path,
      newPath = operation.newPath,
      newIndex = operation.newIndex;
  var _state5 = state,
      document = _state5.document;

  var node = document.assertPath(path);

  // Remove the node from its current parent
  var parent = document.getParent(node.key);
  var isParent = document == parent;
  var index = parent.nodes.indexOf(node);
  parent = parent.removeNode(index);
  document = isParent ? parent : document.updateDescendant(parent);

  // Insert the new node to its new parent
  var target = document.assertPath(newPath);
  var isTarget = document == target;
  target = target.insertNode(newIndex, node);
  document = isTarget ? target : document.updateDescendant(target);

  state = state.merge({ document: document });
  return state;
}

/**
 * Remove mark from text at `offset` and `length` in node by `path`.
 *
 * @param {State} state
 * @param {Object} operation
 * @return {State}
 */

function removeMark(state, operation) {
  var path = operation.path,
      offset = operation.offset,
      length = operation.length,
      mark = operation.mark;
  var _state6 = state,
      document = _state6.document;

  var node = document.assertPath(path);
  node = node.removeMark(offset, length, mark);
  document = document.updateDescendant(node);
  state = state.merge({ document: document });
  return state;
}

/**
 * Remove a node by `path`.
 *
 * @param {State} state
 * @param {Object} operation
 * @return {State}
 */

function removeNode(state, operation) {
  var path = operation.path;
  var _state7 = state,
      document = _state7.document,
      selection = _state7.selection;
  var _selection3 = selection,
      startKey = _selection3.startKey,
      endKey = _selection3.endKey;

  // Preserve previous document

  var prevDocument = document;

  // Update the document
  var node = document.assertPath(path);
  var parent = document.getParent(node.key);
  var index = parent.nodes.indexOf(node);
  var isParent = document == parent;
  parent = parent.removeNode(index);
  document = isParent ? parent : document.updateDescendant(parent);

  function getRemoved(key) {
    if (key === node.key) return node;
    if (node.kind == 'text') return null;
    return node.getDescendant(key);
  }

  // Update the selection, if one of the anchor/focus has been removed
  var startDesc = startKey ? getRemoved(startKey) : null;
  var endDesc = endKey ? getRemoved(endKey) : null;

  if (startDesc) {
    var prevText = prevDocument.getTexts().takeUntil(function (text) {
      return text.key == startKey;
    }).filter(function (text) {
      return !getRemoved(text.key);
    }).last();
    selection = !prevText ? selection.unset() : selection.moveStartTo(prevText.key, prevText.length);
  }
  if (endDesc) {
    // The whole selection is inside the node, we collapse to the previous text node
    if (startKey == endKey) {
      selection = selection.collapseToStart();
    } else {
      var nextText = prevDocument.getTexts().skipUntil(function (text) {
        return text.key == startKey;
      }).slice(1).filter(function (text) {
        return !getRemoved(text.key);
      }).first();

      selection = !nextText ? selection.unset() : selection.moveEndTo(nextText.key, 0);
    }
  }

  state = state.merge({ document: document, selection: selection });
  return state;
}

/**
 * Remove text at `offset` and `length` in node by `path`.
 *
 * @param {State} state
 * @param {Object} operation
 * @return {State}
 */

function removeText(state, operation) {
  var path = operation.path,
      offset = operation.offset,
      length = operation.length;
  var _state8 = state,
      document = _state8.document,
      selection = _state8.selection;
  var _selection4 = selection,
      startKey = _selection4.startKey,
      endKey = _selection4.endKey,
      startOffset = _selection4.startOffset,
      endOffset = _selection4.endOffset;

  var node = document.assertPath(path);

  var rangeOffset = offset + length;

  // Update the document
  node = node.removeText(offset, length);
  document = document.updateDescendant(node);

  // Update the selection
  if (startKey == node.key && startOffset >= rangeOffset) {
    selection = selection.moveStartOffset(-length);
  }
  if (endKey == node.key && endOffset >= rangeOffset) {
    selection = selection.moveEndOffset(-length);
  }

  state = state.merge({ document: document, selection: selection });
  return state;
}

/**
 * Set `properties` on mark on text at `offset` and `length` in node by `path`.
 *
 * @param {State} state
 * @param {Object} operation
 * @return {State}
 */

function setMark(state, operation) {
  var path = operation.path,
      offset = operation.offset,
      length = operation.length,
      mark = operation.mark,
      newMark = operation.newMark;
  var _state9 = state,
      document = _state9.document;

  var node = document.assertPath(path);
  node = node.updateMark(offset, length, mark, newMark);
  document = document.updateDescendant(node);
  state = state.merge({ document: document });
  return state;
}

/**
 * Set `properties` on a node by `path`.
 *
 * @param {State} state
 * @param {Object} operation
 * @return {State}
 */

function setNode(state, operation) {
  var path = operation.path,
      properties = operation.properties;
  var _state10 = state,
      document = _state10.document;

  var node = document.assertPath(path);

  // Deprecate the ability to overwite a node's children.
  if (properties.nodes && properties.nodes != node.nodes) {
    (0, _warn2.default)('Updating a Node\'s `nodes` property via `setNode()` is not allowed. Use the appropriate insertion and removal operations instead. The opeartion in question was:', operation);
    delete properties.nodes;
  }

  // Deprecate the ability to change a node's key.
  if (properties.key && properties.key != node.key) {
    (0, _warn2.default)('Updating a Node\'s `key` property via `setNode()` is not allowed. There should be no reason to do this. The opeartion in question was:', operation);
    delete properties.key;
  }

  node = node.merge(properties);
  document = document.updateDescendant(node);
  state = state.merge({ document: document });
  return state;
}

/**
 * Set `properties` on the selection.
 *
 * @param {State} state
 * @param {Object} operation
 * @return {State}
 */

function setSelection(state, operation) {
  var properties = _extends({}, operation.properties);
  var _state11 = state,
      document = _state11.document,
      selection = _state11.selection;


  if (properties.anchorPath !== undefined) {
    properties.anchorKey = properties.anchorPath === null ? null : document.assertPath(properties.anchorPath).key;
    delete properties.anchorPath;
  }

  if (properties.focusPath !== undefined) {
    properties.focusKey = properties.focusPath === null ? null : document.assertPath(properties.focusPath).key;
    delete properties.focusPath;
  }

  selection = selection.merge(properties);
  selection = selection.normalize(document);
  state = state.merge({ selection: selection });
  return state;
}

/**
 * Split a node by `path` at `offset`.
 *
 * @param {State} state
 * @param {Object} operation
 *   @param {Array} operation.path The path of the node to split
 *   @param {Number} operation.offset (optional) Split using a relative offset
 *   @param {Number} operation.count (optional) Split after `count`
 *   children. Cannot be used in combination with offset.
 * @return {State}
 */

function splitNode(state, operation) {
  var path = operation.path,
      offset = operation.offset,
      count = operation.count;
  var _state12 = state,
      document = _state12.document;


  if (offset === undefined) {
    return state.merge({
      document: document.splitNodeAfter(path, count)
      // No need to update selection
    });
  } else {
    // Update document
    var newDocument = document.splitNode(path, offset);

    // Update selection
    var _state13 = state,
        selection = _state13.selection;
    var _selection5 = selection,
        anchorKey = _selection5.anchorKey,
        anchorOffset = _selection5.anchorOffset,
        focusKey = _selection5.focusKey,
        focusOffset = _selection5.focusOffset;


    var node = document.assertPath(path);
    // The text node that was split
    var splittedText = node.kind == 'text' ? node : node.getTextAtOffset(offset);
    var textOffset = node.kind == 'text' ? offset : offset - node.getOffset(splittedText.key);

    // Should we update the selection ?
    var shouldUpdateAnchor = splittedText.key == anchorKey && textOffset <= anchorOffset;
    var shouldUpdateFocus = splittedText.key == focusKey && textOffset <= focusOffset;
    if (shouldUpdateFocus || shouldUpdateAnchor) {
      // The node next to `node`, resulting from the split
      var secondNode = newDocument.getNextSibling(node.key);
      var secondText = void 0,
          newOffset = void 0;

      if (shouldUpdateAnchor) {
        newOffset = anchorOffset - textOffset;
        secondText = secondNode.kind == 'text' ? secondNode : secondNode.getTextAtOffset(newOffset);
        selection = selection.merge({
          anchorKey: secondText.key,
          anchorOffset: newOffset
        });
      }

      if (shouldUpdateFocus) {
        newOffset = focusOffset - textOffset;
        secondText = secondNode.kind == 'text' ? secondNode : secondNode.getTextAtOffset(newOffset);
        selection = selection.merge({
          focusKey: secondText.key,
          focusOffset: newOffset
        });
      }
    }

    state = state.merge({
      document: newDocument,
      selection: selection
    });
    return state;
  }
}