'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.normalize = normalize;
exports.normalizeDocument = normalizeDocument;
exports.normalizeNodeByKey = normalizeNodeByKey;
exports.normalizeParentsByKey = normalizeParentsByKey;
exports.normalizeSelection = normalizeSelection;

var _normalize = require('../utils/normalize');

var _normalize2 = _interopRequireDefault(_normalize);

var _schema = require('../models/schema');

var _schema2 = _interopRequireDefault(_schema);

var _warn = require('../utils/warn');

var _warn2 = _interopRequireDefault(_warn);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * Normalize the document and selection with a `schema`.
 *
 * @param {Transform} transform
 * @param {Schema} schema
 * @return {Transform}
 */

function normalize(transform, schema) {
  return transform.normalizeDocument(schema).normalizeSelection(schema);
}

/**
 * Normalize the document with a `schema`.
 *
 * @param {Transform} transform
 * @param {Schema} schema
 * @return {Transform}
 */

function normalizeDocument(transform, schema) {
  var state = transform.state;
  var document = state.document;

  return transform.normalizeNodeByKey(document.key, schema);
}

/**
 * Normalize a `node` and its children with a `schema`.
 *
 * @param {Transform} transform
 * @param {Node|String} key
 * @param {Schema} schema
 * @return {Transform}
 */

function normalizeNodeByKey(transform, key, schema) {
  assertSchema(schema);
  key = _normalize2.default.key(key);

  // If the schema has no validation rules, there's nothing to normalize.
  if (!schema.hasValidators) {
    return transform;
  }

  var state = transform.state;
  var document = state.document;

  var node = document.assertNode(key);

  normalizeNodeWith(transform, node, schema);
  return transform;
}

/**
 * Normalize a `node` and its parents with a `schema`.
 *
 * @param {Transform} transform
 * @param {Node|String} key
 * @param {Schema} schema
 * @return {Transform}
 */

function normalizeParentsByKey(transform, key, schema) {
  assertSchema(schema);
  key = _normalize2.default.key(key);

  // If the schema has no validation rules, there's nothing to normalize.
  if (!schema.hasValidators) {
    return transform;
  }

  var state = transform.state;
  var document = state.document;

  var node = document.assertNode(key);

  normalizeParentsWith(transform, node, schema);
  return transform;
}

/**
 * Normalize only the selection.
 *
 * @param {Transform} transform
 * @return {Transform}
 */

function normalizeSelection(transform) {
  var state = transform.state;
  var _state = state,
      document = _state.document,
      selection = _state.selection;

  selection = selection.normalize(document);

  // If the selection is unset, or the anchor or focus key in the selection are
  // pointing to nodes that no longer exist, warn and reset the selection.
  if (selection.isUnset || !document.hasDescendant(selection.anchorKey) || !document.hasDescendant(selection.focusKey)) {
    (0, _warn2.default)('The selection was invalid and was reset to start of the document. The selection in question was:', selection);

    var firstText = document.getFirstText();
    selection = selection.merge({
      anchorKey: firstText.key,
      anchorOffset: 0,
      focusKey: firstText.key,
      focusOffset: 0,
      isBackward: false
    });
  }

  state = state.merge({ selection: selection });
  transform.state = state;
  return transform;
}

/**
 * Normalize a `node` and its children with a `schema`.
 *
 * @param {Transform} transform
 * @param {Node} node
 * @param {Schema} schema
 * @return {Transform}
 */

function normalizeNodeWith(transform, node, schema) {
  // For performance considerations, we will check if the transform was changed.
  var opCount = transform.operations.length;

  // Iterate over its children.
  normalizeChildrenWith(transform, node, schema);

  // Re-find the node reference if necessary.
  if (transform.operations.length != opCount) {
    node = refindNode(transform, node);
  }

  // Now normalize the node itself if it still exists.
  if (node) {
    normalizeNodeOnly(transform, node, schema);
  }

  return transform;
}

/**
 * Normalize a `node` and its parents with a `schema`.
 *
 * @param {Transform} transform
 * @param {Node} node
 * @param {Schema} schema
 * @return {Transform}
 */

function normalizeParentsWith(transform, node, schema) {
  normalizeNodeOnly(transform, node, schema);

  // Normalize went back up to the very top of the document.
  if (node.kind == 'document') {
    return transform;
  }

  // Re-find the node first.
  node = refindNode(transform, node);

  if (!node) {
    return transform;
  }

  var state = transform.state;
  var document = state.document;

  var parent = document.getParent(node.key);

  return normalizeParentsWith(transform, parent, schema);
}

/**
 * Re-find a reference to a node that may have been modified or removed
 * entirely by a transform.
 *
 * @param {Transform} transform
 * @param {Node} node
 * @return {Node}
 */

function refindNode(transform, node) {
  var state = transform.state;
  var document = state.document;

  return node.kind == 'document' ? document : document.getDescendant(node.key);
}

/**
 * Normalize the children of a `node` with a `schema`.
 *
 * @param {Transform} transform
 * @param {Node} node
 * @param {Schema} schema
 * @return {Transform}
 */

function normalizeChildrenWith(transform, node, schema) {
  if (node.kind == 'text') return transform;

  node.nodes.forEach(function (child) {
    normalizeNodeWith(transform, child, schema);
  });

  return transform;
}

/**
 * Normalize a `node` with a `schema`, but not its children.
 *
 * @param {Transform} transform
 * @param {Node} node
 * @param {Schema} schema
 * @return {Transform}
 */

function normalizeNodeOnly(transform, node, schema) {
  var max = schema.rules.length;
  var iterations = 0;

  function iterate(t, n) {
    var failure = n.validate(schema);
    if (!failure) return t;

    var value = failure.value,
        rule = failure.rule;

    // Rule the `normalize` function for the rule with the invalid value.

    rule.normalize(t, n, value);

    // Re-find the node reference, in case it was updated. If the node no longer
    // exists, we're done for this branch.
    n = refindNode(t, n);
    if (!n) return t;

    // Increment the iterations counter, and check to make sure that we haven't
    // exceeded the max. Without this check, it's easy for the `validate` or
    // `normalize` function of a schema rule to be written incorrectly and for
    // an infinite invalid loop to occur.
    iterations++;

    if (iterations > max) {
      throw new Error('A schema rule could not be validated after sufficient iterations. This is usually due to a `rule.validate` or `rule.normalize` function of a schema being incorrectly written, causing an infinite loop.');
    }

    // Otherwise, iterate again.
    return iterate(t, n);
  }

  return iterate(transform, node);
}

/**
 * Assert that a `schema` exists.
 *
 * @param {Schema} schema
 */

function assertSchema(schema) {
  if (schema instanceof _schema2.default) return;

  if (schema == null) {
    throw new Error('You must pass a `schema` object.');
  } else {
    throw new Error('You passed an invalid `schema` object: ' + schema + '.');
  }
}